/* xapian-check.cc: Check consistency of a chert or flint database or table.
 *
 * Copyright 1999,2000,2001 BrightStation PLC
 * Copyright 2002,2003,2004,2005,2006,2007,2008,2009 Olly Betts
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
 * USA
 */

#include <config.h>
#include <iostream>

#include "autoptr.h"

#include "bitstream.h"

#include "chert_check.h"
#include "chert_cursor.h"
#include "chert_table.h"
#include "chert_types.h"

#include "internaltypes.h"

#include "stringutils.h"
#include "utils.h"
#include "valuestats.h"

#include "xapian-check-flint.h"

#include <xapian.h>

#include "safeerrno.h"
#include <sys/types.h>
#include "safesysstat.h"
#include "safefcntl.h"

using namespace std;

#define PROG_NAME "xapian-check"
#define PROG_DESC "Check the consistency of a chert or flint database or table"

// FIXME: We don't currently cross-check wdf between postlist and termlist.
// It's hard to see how to efficiently.  We do cross-check doclens, but that
// "only" requires (4 * last_docid()) bytes.

static void show_usage() {
    cout << "Usage: "PROG_NAME" <chert or flint directory>|<path to btree and prefix> [[t][f][b][v][+]]\n\n"
"The btree(s) is/are always checked - control the output verbosity with:\n"
" t = short tree printing\n"
" f = full tree printing\n"
" b = show bitmap\n"
" v = show stats about B-tree (default)\n"
" + = same as tbv\n"
" e.g. "PROG_NAME" /var/lib/xapian/data/default\n"
"      "PROG_NAME" /var/lib/xapian/data/default/postlist fbv" << endl;
}

static size_t check_chert_table(const char * tablename, string table, int opts,
				vector<Xapian::termcount> & doclens,
				Xapian::docid db_last_docid);

static inline bool
is_user_metadata_key(const string & key)
{
    return key.size() > 1 && key[0] == '\0' && key[1] == '\xc0';
}

int
main(int argc, char **argv)
{
    if (argc > 1 && argv[1][0] == '-') {
	if (strcmp(argv[1], "--help") == 0) {
	    cout << PROG_NAME" - "PROG_DESC"\n\n";
	    show_usage();
	    exit(0);
	}
	if (strcmp(argv[1], "--version") == 0) {
	    cout << PROG_NAME" - "PACKAGE_STRING << endl;
	    exit(0);
	}
    }
    if (argc < 2 || argc > 3) {
	show_usage();
	exit(1);
    }

    int opts = 0;
    const char * opt_string = argv[2];
    if (!opt_string) opt_string = "v";
    for (const char *p = opt_string; *p; ++p) {
	switch (*p) {
	    case 't': opts |= OPT_SHORT_TREE; break;
	    case 'f': opts |= OPT_FULL_TREE; break;
	    case 'b': opts |= OPT_SHOW_BITMAP; break;
	    case 'v': opts |= OPT_SHOW_STATS; break;
	    case '+':
		opts |= OPT_SHORT_TREE | OPT_SHOW_BITMAP | OPT_SHOW_STATS;
		break;
	    default:
		cerr << "option " << opt_string << " unknown\n";
		cerr << "use t,f,b,v and/or + in the option string\n";
		exit(1);
	}
    }

    try {
	vector<Xapian::termcount> doclens;
	size_t errors = 0;
	struct stat sb;
	string dir(argv[1]);
#if 0
	if (stat((dir + "/iamflint").c_str(), &sb) == 0) {
	    // Check a whole flint database directory.
	    try {
		Xapian::Database db = Xapian::Flint::open(dir);
		doclens.reserve(db.get_lastdocid() + 1);
	    } catch (const Xapian::Error & e) {
		// Ignore so we can check a database too broken to open.
		cout << "Database couldn't be opened for reading: "
		     << e.get_description()
		     << "\nContinuing check anyway" << endl;
		++errors;
	    }
	    // This is a flint directory so try to check all the btrees.
	    // Note: it's important to check termlist before postlist so
	    // that we can cross-check the document lengths.
	    const char * tables[] = {
		"record", "termlist", "postlist", "position", "value",
		"spelling", "synonym"
	    };
	    for (const char **t = tables;
		 t != tables + sizeof(tables)/sizeof(tables[0]); ++t) {
		string table(dir);
		table += '/';
		table += *t;
		cout << *t << ":\n";
		if (strcmp(*t, "position") == 0 ||
		    strcmp(*t, "value") == 0 ||
		    strcmp(*t, "spelling") == 0 ||
		    strcmp(*t, "synonym") == 0) {
		    // These are created lazily, so may not exist.
		    if (!file_exists(table + ".DB")) {
			cout << "Lazily created, and not yet used.\n" << endl;
			continue;
		    }
		}
		errors += check_flint_table(*t, table, opts, doclens);
	    }
	} else
#endif
#ifdef XAPIAN_HAS_CHERT_BACKEND
	if (stat((dir + "/iamchert").c_str(), &sb) == 0) {
	    // Check a whole chert database directory.
	    // If we can't read the last docid, set it to its maximum value
	    // to suppress errors.
	    Xapian::docid db_last_docid = static_cast<Xapian::docid>(-1);
	    try {
		Xapian::Database db = Xapian::Chert::open(dir);
		db_last_docid = db.get_lastdocid();
		doclens.reserve(db_last_docid + 1);
	    } catch (const Xapian::Error & e) {
		// Ignore so we can check a database too broken to open.
		cout << "Database couldn't be opened for reading: "
		     << e.get_description()
		     << "\nContinuing check anyway" << endl;
		++errors;
	    }
	    // This is a chert directory so try to check all the btrees.
	    // Note: it's important to check termlist before postlist so
	    // that we can cross-check the document lengths.
	    const char * tables[] = {
		"record", "termlist", "postlist", "position",
		"spelling", "synonym"
	    };
	    for (const char **t = tables;
		 t != tables + sizeof(tables)/sizeof(tables[0]); ++t) {
		string table(dir);
		table += '/';
		table += *t;
		cout << *t << ":\n";
		if (strcmp(*t, "position") == 0 ||
		    strcmp(*t, "spelling") == 0 ||
		    strcmp(*t, "synonym") == 0) {
		    // These are created lazily, so may not exist.
		    if (!file_exists(table + ".DB")) {
			cout << "Lazily created, and not yet used.\n" << endl;
			continue;
		    }
		}
		errors += check_chert_table(*t, table, opts, doclens,
					    db_last_docid);
	    }
	} else
#endif
	{
	    if (stat((dir + "/record_DB").c_str(), &sb) == 0) {
		// Quartz is no longer supported as of Xapian 1.1.0.
		cerr << argv[0] << ": '" << dir << "' is a quartz database.\n"
			"Support for quartz was dropped in Xapian 1.1.0" << endl;
		exit(1);
	    }
	    // Just check a single Btree.  If it ends with "." or ".DB"
	    // already, trim that so the user can do xapian-check on
	    // "foo", "foo.", or "foo.DB".
	    string filename = dir;
	    if (endswith(filename, '.'))
		filename.resize(filename.size() - 1);
	    else if (endswith(filename, ".DB"))
		filename.resize(filename.size() - 3);

	    size_t p = filename.find_last_of('/');
#if defined __WIN32__ || defined __EMX__
	    if (p == string::npos) p = 0;
	    p = filename.find_last_of('\\', p);
#endif
	    if (p == string::npos) p = 0; else ++p;

	    bool flint = false; //!file_exists(filename.substr(0, p) + "iamchert");

	    string tablename;
	    while (p != filename.size()) {
		tablename += tolower(static_cast<unsigned char>(filename[p++]));
	    }

	    if (flint) {
		errors = check_flint_table(tablename.c_str(), filename, opts,
					   doclens);
	    } else {
		// Set the last docid to its maximum value to suppress errors.
		Xapian::docid db_last_docid = static_cast<Xapian::docid>(-1);
		errors = check_chert_table(tablename.c_str(), filename, opts,
					   doclens, db_last_docid);
	    }
	}
	if (errors > 0) {
	    cout << "Total errors found: " << errors << endl;
	    exit(1);
	}
	cout << "No errors found" << endl;
    } catch (const char *error) {
	cerr << argv[0] << ": " << error << endl;
	exit(1);
    } catch (const Xapian::Error &error) {
	cerr << argv[0] << ": " << error.get_description() << endl;
	exit(1);
    } catch (...) {
	cerr << argv[0] << ": Unknown exception" << endl;
	exit(1);
    }
}

static size_t
check_chert_table(const char * tablename, string filename, int,
		  vector<Xapian::termcount> &,
		  Xapian::docid)
{
    filename += '.';

    ChertTable table(tablename, filename, true);
    table.open();
    AutoPtr<ChertCursor> cursor(table.cursor_get());

    unsigned long long leftover = 0;
    unsigned long long blocks = 0;
    struct stat sb;
    if (stat((filename + "DB").c_str(), &sb) == 0) {
	blocks = sb.st_size / table.get_block_size();
	leftover = sb.st_size;
    } else {
	cout << "stat on DB file failed: " << strerror(errno) << endl;
    }
    unsigned long long total_tag = 0;
    unsigned long long total_key = 0;
    unsigned long long entries = 0;

    cursor->find_entry("");
    cursor->next(); // Skip the empty entry.
    unsigned long long key_sizes[256];
    unsigned long long tag_sizes[8192];
    map<size_t, unsigned long long> big_tag_sizes;
    memset(key_sizes, 0, sizeof(key_sizes));
    memset(tag_sizes, 0, sizeof(tag_sizes));
    while (!cursor->after_end()) {
	++entries;
	cursor->read_tag(true); // Check the compressed size.
	size_t tag_size = cursor->current_tag.size();
	unsigned long long old_total = total_tag;
	total_tag += tag_size;
	if (total_tag < old_total) cout << "overflowed tag" << endl;
	if (tag_size >= sizeof(tag_sizes) / sizeof(*tag_sizes)) {
	    ++big_tag_sizes[tag_size];
	} else {
	    ++tag_sizes[tag_size];
	}
	size_t key_size = cursor->current_key.size();
	old_total = total_key;
	total_key += key_size;
	if (total_key < old_total) cout << "overflowed key" << endl;
	if (key_size >= sizeof(key_sizes) / sizeof(*key_sizes)) {
	    cout << "key longer than expected" << endl;
	    key_size = sizeof(key_sizes) / sizeof(*key_sizes) - 1;
	}
	++key_sizes[key_size];
	cursor->next();
    }

    cout << tablename << " " << filename << endl;
    double dentries = entries;
    cout << " total blocks " << blocks << "\n";
    cout << " entries " << entries << endl;
    cout << " table.get_entry_count() " << table.get_entry_count() << endl;
    cout << total_key/dentries << " key " << total_key << " bytes\n";
    leftover -= total_key;
    size_t i;
    for (i = 0; i < sizeof(key_sizes) / sizeof(*key_sizes); ++i) {
	if (key_sizes[i]) {
	    cout << "key len " << i << " seen " << key_sizes[i] << " times\n";
	}
    }
    cout << endl;
    cout << total_tag/dentries << " tag " << total_tag << " bytes\n";
    for (i = 0; i < sizeof(tag_sizes) / sizeof(*tag_sizes); ++i) {
	if (tag_sizes[i]) {
	    cout << "tag len " << i << " seen " << tag_sizes[i] << " times\n";
	}
    }
    map<size_t, unsigned long long>::const_iterator j;
    for (j = big_tag_sizes.begin(); j != big_tag_sizes.end(); ++j) {
	cout << "tag len " << j->first << " seen " << j->second << " times\n";
    }
    cout << endl;
    leftover -= total_tag;
    unsigned per_item = 2 + 1 + 2 + 2 + 2;
    cout << per_item << " per item overhead (" << entries * per_item << ")\n";
    leftover -= entries * per_item;
    cout << endl;
    unsigned long count[256];
    memset(count, 0, 256 * sizeof(unsigned long));
    int fd = open((filename + "DB").c_str(), O_RDONLY);
    off_t pos = 0;
    off_t block_size = table.get_block_size();
    unsigned long long dead_space = 0;
    unsigned long items = 0;
    while (true) {
	ssize_t res;
	unsigned char buf[8192];
again:
	unsigned READ = 11; //8192;
	res = pread(fd, &buf, READ, pos);
	if (res <= 0) {
	    if (res == 0) break;
	    if (errno == EINTR) continue;
	    cout << "pread error " << strerror(errno) << endl;
	} else if (res != READ) {
	    cout << "short! " << res << endl;
	    goto again;
	} else {
	    unsigned char level = buf[4];
	    ++count[level];
	    if (level == 0) {
		dead_space += getint2(buf, 7);
		unsigned items_here = (getint2(buf, 9) - 11) / 2;
		items += items_here;
	    }
	}
	pos += block_size;
    }
    unsigned long long non_leaf_blocks = 0;
    for (size_t level = 0; level < 256; ++level) {
	if (count[level] == 0) continue;
	cout << " level " << level << " has " << count[level] << " blocks\n";
	if (level) non_leaf_blocks += count[level];
    }
    cout << (non_leaf_blocks * block_size)/dentries << " branch blocks = " << non_leaf_blocks;
    cout << " (" << non_leaf_blocks * block_size << " bytes)\n";
    leftover -= non_leaf_blocks * block_size;
    cout << " leaf blocks " << count[0] << endl;
    unsigned long long block_overhead = (4 + 1 + 2 + 2 + 2) * count[0];
    cout << block_overhead/dentries << " (block header " << block_overhead << ")\n";
    leftover -= block_overhead;
    cout << dead_space/dentries;
    cout << " unused space at end of leaf blocks = " << dead_space << " bytes";
    cout << " (" << dead_space/(double)count[0] << " per leaf block)" << endl;
    cout << " " << entries << " entries + " << items - entries << " extra items from splitting  = " << items << " total items\n";
    leftover -= dead_space;
    cout << leftover/dentries;
    cout << " space used by splitting large tags into multiple items = " << leftover << " bytes\n";
    cout << " (" << leftover/(double)count[0] << " per leaf block)" << endl;
    cout << endl;
    return 0;
}
